"""Utility functions for logging."""
import functools


def log_entrance_exit(func):
    """
    Decorator to log the entrance and exit of method calls.

    This will make use of the logger instance passed via self.  When using
    functools.wraps, self will be the first index in args.
    See self = args[0] comment below.

    Args:
        func: The function to be wrapped.

    Returns:
        Wrapped function.

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # self should be first argument.
        func_self = args[0]
        if len(args) > 1:
            func_self.logger.info('Entering %s: %s, %s', func.__name__, args[1:], kwargs)
        else:
            func_self.logger.info('Entering %s: %s', func.__name__, kwargs)

        # Evaluate wrapped function.
        f_result = func(*args, **kwargs)

        if len(args) > 1:
            func_self.logger.info('Exiting %s: %s, %s', func.__name__, args[1:], kwargs)
        else:
            func_self.logger.info('Exiting %s: %s', func.__name__, kwargs)
        return f_result

    return wrapper



"""Mixins and core functionality."""
import json
import os
from contextlib import closing
from io import BytesIO
from itertools import repeat
from math import ceil
from multiprocessing.pool import ThreadPool

import numpy as np
from PIL import Image
from helios import BASE_API_URL
from helios.utilities import logging_utils


class SDKCore(object):
    """
    Core class for Python interface to Helios Core APIs.

    This class must be inherited by any additional Core API classes.
    """
    _base_api_url = BASE_API_URL

    @staticmethod
    def parse_query_inputs(input_dict):
        # Check for unique case: sensors
        if 'sensors' in input_dict.keys():
            query_str = input_dict.pop('sensors') + '&'
        else:
            query_str = ''

        # Parse input_dict into a str
        for key in input_dict.keys():
            if isinstance(input_dict[key], (list, tuple)):
                query_str += (str(key) +
                              '=' +
                              ','.join([str(x) for x in input_dict[key]]) +
                              '&')
            else:
                query_str += (str(key) + '=' + str(input_dict[key]) + '&')
        query_str = query_str[:-1]

        return query_str

    @staticmethod
    def check_headers_for_dud(header_data):
        if 'x-amz-meta-helios' in header_data:
            meta_block = json.loads(header_data['x-amz-meta-helios'])
            if meta_block['isOutcast'] or meta_block['isDud'] or meta_block['isFrozen']:
                return True
        return False

    @property
    def base_api_url(self):
        return self._base_api_url

    @base_api_url.setter
    def base_api_url(self, value):
        raise AttributeError('Access to base_api_url is restricted.')


class IndexMixin(object):
    @logging_utils.log_entrance_exit
    def index(self, **kwargs):
        max_skip = 4000
        limit = kwargs.get('limit', 100)
        skip = kwargs.get('skip', 0)

        # Establish all queries.
        params_str = self.parse_query_inputs(kwargs)
        queries = []
        for i in range(skip, max_skip, limit):
            if i + limit > max_skip:
                temp_limit = max_skip - i
            else:
                temp_limit = limit

            query_str = '{}/{}?{}&limit={}&skip={}'.format(self.base_api_url,
                                                           self.core_api,
                                                           params_str,
                                                           temp_limit,
                                                           i)

            queries.append(query_str)

        # Do first query to find total number of results to expect.
        initial_resp = self.request_manager.get(queries.pop(0)).json()

        try:
            total = initial_resp['properties']['total']
        except KeyError:
            total = initial_resp['total']

        # Warn the user when truncation occurs. (max_skip is hit)
        if total > max_skip:
            # Log truncation warning
            self.logger.warning('Maximum skip level. Truncated results for: %s',
                                kwargs)

        # Get number of results in initial query.
        try:
            n_features = len(initial_resp['features'])
        except KeyError:
            n_features = len(initial_resp['results'])

        # If only one query was necessary, return immediately.
        if n_features < limit:
            return [initial_resp]

        # Determine number of iterations that will be needed.
        n_queries_needed = int(ceil((total - skip) / float(limit))) - 1
        queries = queries[0:n_queries_needed]

        # Log number of queries required.
        self.logger.info('%s index queries required for: %s', n_queries_needed,
                         kwargs)

        # Create thread pool and get results
        num_threads = min(self.max_threads, n_queries_needed)
        if num_threads > 1:
            with closing(ThreadPool(num_threads)) as thread_pool:
                results = thread_pool.map(self.__index_worker, queries)
        else:
            results = [self.__index_worker(queries[0])]

        # Put initial query back in list.
        results.insert(0, initial_resp)

        return results

    def __index_worker(self, args):
        query_str = args

        # Perform query
        resp = self.request_manager.get(query_str)

        return resp.json()


class ShowMixin(object):
    @logging_utils.log_entrance_exit
    def show(self, id_var, **kwargs):
        params_str = self.parse_query_inputs(kwargs)
        query_str = '{}/{}/{}?{}'.format(self.base_api_url, self.core_api,
                                         id_var, params_str)

        resp = self.request_manager.get(query_str)

        return resp.json()


class ShowImageMixin(object):
    @logging_utils.log_entrance_exit
    def show_image(self, id_var, samples, check_for_duds=True):
        # Force iterable
        if not isinstance(samples, (list, tuple)):
            samples = [samples]
        n_samples = len(samples)

        # Get number of threads
        num_threads = min(self.max_threads, n_samples)

        # Process urls.
        if num_threads > 1:
            with closing(ThreadPool(num_threads)) as thread_pool:
                results = thread_pool.map(self.__show_image_worker,
                                          zip(repeat(id_var),
                                              samples,
                                              repeat(check_for_duds)))
        else:
            results = [self.__show_image_worker((id_var, samples[0],
                                                 check_for_duds))]

        # Remove errors, if they exist
        results = [x for x in results if x != -1]

        # Determine how many were successful
        n_data = len(results)
        message = 'showImage({} out of {} successful)'.format(n_data, n_samples)

        if n_data == 0:
            self.logger.error(message)
            return -1
        elif n_data < n_samples:
            self.logger.warning(message)
        else:
            self.logger.info(message)

        return results

    def __show_image_worker(self, args):
        id_var, data, check_for_duds = args

        query_str = '{}/{}/{}/images/{}'.format(self.base_api_url,
                                                self.core_api,
                                                id_var,
                                                data)

        try:
            resp = self.request_manager.get(query_str)
            redirect_url = resp.url[0:resp.url.index('?')]
        except Exception:
            return -1

        # Check header for dud statuses.
        if check_for_duds:
            try:
                # Redirect URLs do not use api credentials
                resp2 = self.request_manager.head(redirect_url,
                                                  use_api_cred=False)
            except Exception:
                return -1

            if self.check_headers_for_dud(resp2.headers):
                self.logger.info('showImage query returned dud image: %s',
                                 query_str)
                return None

        return redirect_url


class DownloadImagesMixin(object):
    @logging_utils.log_entrance_exit
    def download_images(self, urls, out_dir=None, return_image_data=False):
        """Download images from URLs.

        Args:
            urls (str or sequence of strs): Image URLs to download from.
            out_dir (str, optional): Output directory to save images to.
                Defaults to None.
            return_image_data (bool, optional): If True, image data will be
                read into a Numpy ndarray and returned. Defaults to False.

        Returns:
            sequence of ndarrays or None: Image data if return_image_data is
            True or None otherwise.

        """
        # Force iterable
        if not isinstance(urls, (list, tuple)):
            urls = [urls]
        n_urls = len(urls)

        if out_dir is not None:
            if not os.path.exists(out_dir):
                os.mkdir(out_dir)

        # Create thread pool
        num_threads = min(self.max_threads, n_urls)
        if num_threads > 1:
            with closing(ThreadPool(num_threads)) as thread_pool:
                data = thread_pool.map(self.__download_images_worker,
                                       zip(urls, repeat(out_dir),
                                           repeat(return_image_data)))
        else:
            data = [self.__download_images_worker((urls[0],
                                                   out_dir,
                                                   return_image_data))]

        # Remove errors, if the exist
        data = [x for x in data if isinstance(x, np.ndarray) or x != -1]

        # Determine how many were successful
        n_data = len(data)
        message = 'downloadImages({} out of {} successful)'.format(n_data, n_urls)

        if n_data == 0:
            self.logger.error(message)
            return -1
        elif n_data < n_urls:
            self.logger.warning(message)
        else:
            self.logger.info(message)

        if return_image_data:
            return data

    def __download_images_worker(self, args):
        url, out_dir, return_image_data = args

        try:
            resp = self.request_manager.get(url, use_api_cred=False)
        except Exception:
            return -1

        # Read image from response
        img = Image.open(BytesIO(resp.content))

        # Write image to file.
        if out_dir is not None:
            _, tail = os.path.split(url)
            out_file = os.path.join(out_dir, tail)
            img.save(out_file)

        # Read and return image data.
        if return_image_data:
            return np.array(img)



from .mixins import DownloadImagesMixin
from .mixins import IndexMixin
from .mixins import SDKCore
from .mixins import ShowImageMixin
from .mixins import ShowMixin
from .request_manager import RequestManager



"""Helper functions for paths and URLs."""
import hashlib
import os
from datetime import datetime

# Python 2 to 3 fix.
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse


def parse_time(data):
    """
    Parse time from a URL or image name.

    Args:
        data (str): Image URL or name.
    Returns:
        datetime: The parsed time as a datetime object.

    """
    time_string = str(os.path.splitext(os.path.split(data)[-1])[0].split('_')[-1])
    time_stamp = datetime.strptime(time_string, '%Y%m%d%H%M%S%f')
    return time_stamp


def parse_camera(data):
    """
    Parse camera name from a URL or image name.

    Args:
        data (str): Image URL or name.
    Returns:
        str: Camera name.

    """
    split1_rev = os.path.splitext(os.path.split(data)[-1])[0][::-1]
    split2 = split1_rev[split1_rev.index('_') + 1:][::-1]
    md5_str = hashlib.md5(split2[split2.index('-') + 1:].encode('utf-8')).hexdigest()

    if split2[0:4] == md5_str[0:4]:
        return split2[5:]

    return split2


def parse_image_name(url):
    """
    Parse image name from a URL.

    Args:
        url (str): Image URL.
    Returns:
        str: Image name.

    """
    return os.path.split(url)[-1]


def parse_url(url):
    """
    Parse a URL into its components.

    Args:
        url (str): Image URL.
    Returns:
        obj: Parsed URL.

    """
    return urlparse(url)



"""
SDK for the Helios Collections API.

Methods are meant to represent the core functionality in the developer
documentation.  Some may have additional functionality for convenience.

"""
import hashlib
import logging
from contextlib import closing
from itertools import repeat
from multiprocessing.pool import ThreadPool

from helios.core import SDKCore, IndexMixin, ShowMixin, ShowImageMixin, \
    DownloadImagesMixin, RequestManager
from helios.utilities import logging_utils


class Collections(DownloadImagesMixin, ShowImageMixin, ShowMixin, IndexMixin, SDKCore):
    """
    The Collections API allows users to group and organize individual image
    frames.

    Collections are intended to be short-lived resources and will be accessible
    for 90 days from the time the collection was created. After that time
    period has expired, the collection and all associated imagery will be
    removed from the system.

    """
    core_api = 'collections'
    max_threads = 32

    def __init__(self):
        self.request_manager = RequestManager(pool_maxsize=self.max_threads)
        self.logger = logging.getLogger(__name__)

    def index(self, **kwargs):
        """
        Return a list of collections matching the provided spatial, text, or
        metadata filters.

        The maximum skip value is 4000. If this is reached, truncated results
        will be returned. You will need to refine your query to avoid this.

        Args:
            **kwargs: Any keyword arguments found in the documentation.

        Returns:
             list: GeoJSON feature collections.

        """
        return super(Collections, self).index(**kwargs)

    def show(self, collection_id, limit=200, marker=None):
        """
        Return the attributes and image list for a single collection.

        The results will also contain image names available in the collection.
        These are limited to a maximum of 200 per query.

        Args:
            collection_id (str): Collection ID.
            limit (int, optional): Number of image names to be returned with
                each response. Defaults to 200. Max value of 200 is allowed.
            marker (str, optional): Pagination marker. If the marker is an
                exact match to an existing image, the next image after the
                marker will be the first image returned. Therefore, for normal
                linked list pagination, specify the last image name from the
                current response as the marker value in the next request.
                Partial file names may be specified, in which case the first
                matching result will be the first image returned.

        Returns:
            dict: Dictionary containing collection attributes and image list.

        """
        return super(Collections, self).show(collection_id,
                                             limit=limit,
                                             marker=marker)

    @logging_utils.log_entrance_exit
    def create(self, name, description, tags=None):
        """
        Create a new collection.

        Args:
            name (str): Display name for the collection.
            description (str): Description for the collection.
            tags (str or sequence of strs, optional): Optional comma-delimited
                keyword tags to be added to the collection.

        Returns:
            str: New collection ID.

        """
        # need to strip out the Bearer to work with a POST for collections
        post_token = self.request_manager.auth_token['value'].replace('Bearer ', '')

        # handle more than one tag
        if isinstance(tags, (list, tuple)):
            tags = ','.join(tags)

        # Compose parms block
        parms = {'name': name, 'description': description}
        if tags is not None:
            parms['tags'] = tags
        parms['access_token'] = post_token

        header = {'name': 'Content-Type',
                  'value': 'application/x-www-form-urlencoded'}

        post_url = '{}/{}'.format(self.base_api_url, self.core_api)

        resp = self.request_manager.post(post_url, headers=header, data=parms).json()

        return resp['collection_id']

    @logging_utils.log_entrance_exit
    def update(self, collections_id, name=None, description=None, tags=None):
        """
        Update a collection.

        Args:
            collections_id (str): Collection ID.
            name (str, optional): Name to be changed to.
            description (str, optional): Description to be changed to.
            tags (str or sequence of strs, optional): Optional comma-delimited
                keyword tags to be changed to.

        """
        if name is None and description is None and tags is None:
            raise ValueError('Update requires at least one named argument.')

        # need to strip out the Bearer to work with a PATCH for collections
        patch_token = self.request_manager.auth_token['value'].replace('Bearer ', '')

        # handle more than one tag
        if isinstance(tags, (list, tuple)):
            tags = ','.join(tags)

        # Compose parms block
        parms = {}
        if name is not None:
            parms['name'] = name
        if description is not None:
            parms['description'] = description
        if tags is not None:
            parms['tags'] = tags
        parms['access_token'] = patch_token

        header = {'name': 'Content-Type',
                  'value': 'application/x-www-form-urlencoded'}

        patch_url = '{}/{}/{}'.format(self.base_api_url,
                                      self.core_api,
                                      collections_id)

        self.request_manager.patch(patch_url, headers=header, data=parms)

    @logging_utils.log_entrance_exit
    def images(self, collection_id, camera=None, old_flag=False):
        """
        Returns all image names in a given collection.

        When using the optional camera input parameter only images from that
        camera will be returned.

        Args:
            collection_id (str): Collection ID.
            camera (str, optional): Camera ID to be found.
            old_flag (bool, optional): Flag for finding old format image names.
                When True images that do not contain md5 hashes at the start of
                their name will be found.

        Returns:
            sequence of strs: Image names.

        """
        max_limit = 200
        mark_img = ''

        if camera is not None:
            md5_str = hashlib.md5(camera.encode('utf-8')).hexdigest()
            if not old_flag:
                camera = md5_str[0:4] + '-' + camera
            mark_img = camera

        good_images = []
        while True:
            results = self.show(collection_id,
                                limit=max_limit,
                                marker=mark_img)

            # Gather images.
            images_found = results['images']

            if camera is not None:
                imgs_found_temp = [x for x in images_found if x.split('_')[0] == camera]
            else:
                imgs_found_temp = images_found

            if not imgs_found_temp:
                break

            good_images.extend(imgs_found_temp)
            if len(imgs_found_temp) < len(images_found):
                break
            else:
                mark_img = good_images[-1]

        return good_images

    def show_image(self, collection_id, image_names, check_for_duds=False):
        """
        Return image URLs from a collection.

        Args:
            collection_id (str): Collection ID.
            image_names (str or sequence of strs): Image names.
            check_for_duds (bool, optional): Flag for the removal of dud
                images. Defaults to False.

        Returns:
            sequence of strs: Image URLs.

        """
        return super(Collections, self).show_image(
            collection_id, image_names, check_for_duds=check_for_duds)

    @logging_utils.log_entrance_exit
    def add_image(self, collection_id, data):
        """
        Add images to a collection.

        Args:
            collection_id (str): Collection ID.
            data (dict or sequence of dicts): Data containing any of these
                payloads (camera_id), (camera_id, time), (observation_id),
                (collection_id, image). E.g. data =
                [{'camera_id': 'cam_01', time: '2017-01-01T00:00:000Z'}]

        Returns:
            If errors do occur then the data that caused the errors will be
            returned.

        """
        assert isinstance(data, (list, tuple, dict))

        # Force iterable.
        if isinstance(data, dict):
            data = [data]

        n_images = len(data)

        # Get number of threads
        num_threads = min(self.max_threads, n_images)

        # Process data.
        if num_threads > 1:
            with closing(ThreadPool(num_threads)) as thread_pool:
                results = thread_pool.map(self.__add_image_worker,
                                          zip(repeat(collection_id), data))
        else:
            results = self.__add_image_worker((collection_id, data[0]))

        # Extract failures.
        failures = [y for x, y in zip(results, data) if x == -1]

        # Determine how many were successful
        n_successful = n_images - len(failures)
        message = 'addImage({} out of {} successful)'.format(n_successful, n_images)

        if n_successful == 0:
            self.logger.error(message)
            return -1
        elif n_successful < n_images:
            self.logger.warning(message)
        else:
            self.logger.info(message)

        return failures

    def __add_image_worker(self, args):
        collection_id, payload = args

        # need to strip out the Bearer to work with a POST for collections
        post_token = self.request_manager.auth_token['value'].replace('Bearer ', '')

        # Compose post request
        parms = {'access_token': post_token}
        parms.update(payload)

        header = {'name': 'Content-Type',
                  'value': 'application/x-www-form-urlencoded'}
        post_url = '{}/collections/{}/images'.format(self.base_api_url,
                                                     collection_id)

        try:
            self.request_manager.post(post_url, headers=header, data=parms)
        except Exception:
            return -1

    @logging_utils.log_entrance_exit
    def remove_image(self, collection_id, names):
        """
        Remove images from a collection.

        Args:
            collection_id (str): Collection ID.
            names (str or sequence of strs): List of image names to be removed.

        Returns:
            If errors do occur then the data that caused the errors will be
            returned.

        """
        # Force iterable
        if not isinstance(names, (list, tuple)):
            names = [names]
        n_names = len(names)

        # Get number of threads
        num_threads = min(self.max_threads, n_names)

        # Process urls.
        if num_threads > 1:
            with closing(ThreadPool(num_threads)) as thread_pool:
                results = thread_pool.map(self.__remove_image_worker,
                                          zip(repeat(collection_id), names))
        else:
            results = [self.__remove_image_worker((collection_id, names[0]))]

        # Extract failures.
        failures = [y for x, y in zip(results, names) if x == -1]

        # Determine how many were successful
        n_successful = n_names - len(failures)
        message = 'removeImage({} out of {} successful)'.format(n_successful, n_names)

        if n_successful == 0:
            self.logger.error(message)
            return -1
        elif n_successful < n_names:
            self.logger.warning(message)
        else:
            self.logger.info(message)

        return failures

    def __remove_image_worker(self, args):
        coll_id, img_name = args

        query_str = '{}/{}/{}/images/{}'.format(self.base_api_url,
                                                self.core_api,
                                                coll_id,
                                                img_name)

        try:
            self.request_manager.delete(query_str)
        except Exception:
            return -1

    @logging_utils.log_entrance_exit
    def copy(self, collection_id, new_name):
        """
        Copy a collection and its contents to a new collection.

        Args:
            collection_id (str): Collection ID.
            new_name (str): New collection name.

        Returns:
            str: New collection ID.

        """
        # Get the collection metadata that needs to be copied.
        query_str = '{}/{}/{}'.format(self.base_api_url,
                                      self.core_api,
                                      collection_id)
        resp_json = self.request_manager.get(query_str).json()

        # Get the images that exist in the collection.
        image_names = self.images(collection_id)['images']

        # Add images to new collection.
        data = [{'collection_id': collection_id, 'image': x} for x in image_names]

        # Create new collection.
        output = self.create(new_name, resp_json['description'], resp_json['tags'])
        new_id = output['collection_id']

        # Add images to new collection.
        _ = self.add_image(new_id, data)

        return new_id



"""Request manager for all the various components of the Helios SDK."""
import logging

import requests

from helios import AUTH_TOKEN

MAX_RETRIES = 3
TIMEOUT = 5
SSL_VERIFY = True


class RequestManager(object):
    """Manages all API requests."""
    _auth_token = AUTH_TOKEN

    def __init__(self, pool_maxsize=32):
        # Initialize logger
        self.logger = logging.getLogger(__name__)

        # Create API session with authentication credentials
        self.api_session = requests.Session()
        self.api_session.headers.update({AUTH_TOKEN['name']: AUTH_TOKEN['value']})
        self.api_session.verify = SSL_VERIFY
        self.api_session.mount('https://', requests.adapters.HTTPAdapter(
            pool_maxsize=pool_maxsize, max_retries=MAX_RETRIES))

        # Create bare session without credentials
        self.session = requests.Session()
        self.session.verify = SSL_VERIFY
        self.session.mount('https://', requests.adapters.HTTPAdapter(
            pool_maxsize=pool_maxsize, max_retries=MAX_RETRIES))

    @property
    def auth_token(self):
        """Access to authentication token."""
        return self._auth_token

    @auth_token.setter
    def auth_token(self, value):
        raise AttributeError('Access to auth_token is restricted.')

    def __del__(self):
        self.api_session.close()
        self.session.close()

    def __request(self, query, use_api_cred=True, request_type='get',
                  **kwargs):

        query = query.replace(' ', '+')
        kwargs['timeout'] = kwargs.get('timeout', TIMEOUT)

        # Alias the required session
        if use_api_cred:
            sess_alias = self.api_session
        else:
            sess_alias = self.session

        # Perform query and raise exceptions
        try:
            if request_type == 'get':
                resp = sess_alias.get(query, **kwargs)
            elif request_type == 'post':
                resp = sess_alias.post(query, **kwargs)
            elif request_type == 'head':
                resp = sess_alias.head(query, **kwargs)
            elif request_type == 'delete':
                resp = sess_alias.delete(query, **kwargs)
            elif request_type == 'patch':
                resp = sess_alias.patch(query, **kwargs)
            else:
                raise ValueError('Unsupported query of type: {}'.format(request_type))
            resp.raise_for_status()
        # Log exceptions and return
        except requests.exceptions.HTTPError:
            self.logger.error('HTTPError %s: %s', resp.status_code, query)
            raise
        except Exception as e:
            self.logger.error('Error: %s', str(e))
            raise

        return resp

    def get(self, query, use_api_cred=True, **kwargs):
        """
        Perform get request.

        Args:
            query (str): URL string for query.
            use_api_cred (bool, optional): Flag to use API credentials for
                query. Defaults to True.
            **kwargs: Any additional keyword argument for requests.

        Returns:
            Request response.

        """
        return self.__request(query,
                              use_api_cred=use_api_cred,
                              request_type='get',
                              **kwargs)

    def post(self, query, use_api_cred=True, **kwargs):
        """
        Perform post request.

        Args:
            query (str): URL string for query.
            use_api_cred (bool, optional): Flag to use API credentials for
                query. Defaults to True.
            **kwargs: Any additional keyword argument for requests.

        Returns:
            Request response.

        """
        return self.__request(query,
                              use_api_cred=use_api_cred,
                              request_type='post',
                              **kwargs)

    def head(self, query, use_api_cred=True, **kwargs):
        """
        Perform head request.

        Args:
            query (str): URL string for query.
            use_api_cred (bool, optional): Flag to use API credentials for
                query. Defaults to True.
            **kwargs: Any additional keyword argument for requests.

        Returns:
            Request response.

        """
        return self.__request(query,
                              use_api_cred=use_api_cred,
                              request_type='head',
                              **kwargs)

    def delete(self, query, use_api_cred=True, **kwargs):
        """
        Perform delete request.

        Args:
            query (str): URL string for query.
            use_api_cred (bool, optional): Flag to use API credentials for
                query. Defaults to True.
            **kwargs: Any additional keyword argument for requests.

        Returns:
            Request response.

        """
        return self.__request(query,
                              use_api_cred=use_api_cred,
                              request_type='delete',
                              **kwargs)

    def patch(self, query, use_api_cred=True, **kwargs):
        """
        Perform patch request.

        Args:
            query (str): URL string for query.
            use_api_cred (bool, optional): Flag to use API credentials for
                query. Defaults to True.
            **kwargs: Any additional keyword argument for requests.

        Returns:
            Request response.

        """
        return self.__request(query,
                              use_api_cred=use_api_cred,
                              request_type='patch',
                              **kwargs)



from helios.session.logger_setup import configure
from helios.session.token_manager import TokenManager

# Configure logger.
configure()

# Get authentication token and API URL.
AUTH_TOKEN, BASE_API_URL = TokenManager().start_session()

from . import core
from . import utilities
from .alerts import Alerts
from .cameras import Cameras
from .collections import Collections
from .observations import Observations

# Cleanup
del configure
del TokenManager

__version__ = '2.0.0'



"""
SDK for the Helios Alerts API.

Methods are meant to represent the core functionality in the developer
documentation.  Some may have additional functionality for convenience.

"""
import logging

from helios.core import SDKCore, IndexMixin, ShowMixin, RequestManager


class Alerts(ShowMixin, IndexMixin, SDKCore):
    """
    Helios alerts provide real-time severe weather alerts.

    National Weather Service
        Severe weather alerts for the United States are provided by the
        National Weather Service. These alerts cover events such as Flood
        Warnings, Severe Thunderstorm Warnings, and Special Weather Statements.

    Helios
        Alerts generated by Helios are based on the sensor measurements from
        the Observations API. These alerts represent regional areas with a high
        detection confidence and currently include: Road Wetness Watch, Poor
        Visibility Watch, and Heavy Precip Watch.

    """

    core_api = 'alerts'
    max_threads = 32

    def __init__(self):
        self.request_manager = RequestManager(pool_maxsize=self.max_threads)
        self.logger = logging.getLogger(__name__)

    def index(self, **kwargs):
        """
        Return a list of alerts matching the provided spatial, text, or
        metadata filters.

        The maximum skip value is 4000. If this is reached, truncated results
        will be returned. You will need to refine your query to avoid this.

        Args:
            **kwargs: Any keyword arguments found in the documentation.

        Returns:
             list: GeoJSON feature collections.

        """
        return super(Alerts, self).index(**kwargs)

    def show(self, alert_id):
        """
        Return the attributes for a single alert.

        Args:
            alert_id (str): Alert ID.

        Returns:
            dict: GeoJSON feature.

        """
        return super(Alerts, self).show(alert_id)



"""
SDK for the Helios Cameras API.

Methods are meant to represent the core functionality in the developer
documentation.  Some may have additional functionality for convenience.

"""
import logging

from dateutil.parser import parse
from helios.core import SDKCore, ShowMixin, ShowImageMixin, IndexMixin, \
    DownloadImagesMixin, RequestManager
from helios.utilities import logging_utils


class Cameras(DownloadImagesMixin, ShowImageMixin, ShowMixin, IndexMixin,
              SDKCore):
    """The Cameras API provides access to all cameras in the Helios Network."""

    core_api = 'cameras'
    max_threads = 32

    def __init__(self):
        self.request_manager = RequestManager(pool_maxsize=self.max_threads)
        self.logger = logging.getLogger(__name__)

    def index(self, **kwargs):
        """
        Return a list of cameras matching the provided spatial, text, or
        metadata filters.

        The maximum skip value is 4000. If this is reached, truncated results
        will be returned. You will need to refine your query to avoid this.

        Args:
            **kwargs: Any keyword arguments found in the documentation.

        Returns:
             list: GeoJSON feature collections.

        """
        return super(Cameras, self).index(**kwargs)

    def show(self, camera_id):
        """
        Return the attributes for a single alert.

        Args:
            camera_id (str): Camera ID.

        Returns:
            dict: GeoJSON feature.

        """
        return super(Cameras, self).show(camera_id)

    @logging_utils.log_entrance_exit
    def images(self, camera_id, start_time, limit=500):
        """
        Return the image times available for a given camera in the media cache.

        The media cache contains all recent images archived by Helios, either
        for internal analytics or for end user recording purposes.

        Args:
            camera_id (str): Camera ID.
            start_time (str): Starting image timestamp, specified in UTC as an
                ISO 8601 string (e.g. 2014-08-01 or 2014-08-01T12:34:56.000Z).
            limit (int, optional): Number of images to be returned, up to a max
                of 500. Defaults to 500.


        Returns:
            sequence of strs: Image times.

        """
        query_str = '{}/{}/{}/images?time={}&limit={}'.format(self.base_api_url,
                                                              self.core_api,
                                                              camera_id,
                                                              start_time,
                                                              limit)

        resp = self.request_manager.get(query_str).json()

        return resp['times']

    @logging_utils.log_entrance_exit
    def images_range(self, camera_id, start_time, end_time, limit=500):
        """
        Return image times available in a given time range.

        The media cache contains all recent images archived by Helios, either
        for internal analytics or for end user recording purposes.

        Args:
            camera_id (str): Camera ID.
            start_time (str): Starting image timestamp, specified in UTC as an
                ISO 8601 string (e.g. 2014-08-01 or 2014-08-01T12:34:56.000Z).
            end_time (str): Ending image timestamp, specified in UTC as an
                ISO 8601 string (e.g. 2014-08-01 or 2014-08-01T12:34:56.000Z).
            limit (int, optional): Number of images to be returned, up to a max
                of 500.  Defaults to 500.

        Returns:
            sequence of strs: Image times.

        """
        end_time = parse(end_time).utctimetuple()
        image_times = []
        while True:
            times = self.images(camera_id, start_time, limit=limit)

            # If no times exist, break and return.
            if len(times) == 0:
                break

            first = parse(times[0]).utctimetuple()
            last = parse(times[-1]).utctimetuple()

            if first > end_time:
                break

            if len(times) == 1:
                image_times.extend(times)
                break

            # the last image is still newer than our end time, keep looking
            if last < end_time:
                image_times.extend(times)
                start_time = times[-1]
                continue
            else:
                good_times = [x for x in times if parse(x).utctimetuple() < end_time]
                image_times.extend(good_times)
                break

        return image_times

    def show_image(self, camera_id, times):
        """
        Return a single image from the media cache.

        The media cache contains all recent images archived by Helios, either
        for internal analytics or for end user recording purposes.

        Args:
            camera_id (str): Camera ID.
            times (str or sequence of strs): Image times, specified in UTC as an ISO 8601
                string (e.g. 2017-08-01 or 2017-08-01T12:34:56.000Z). The
                image with the closest matching timestamp will be returned.

        Returns:
            sequence of strs: Image URLs.

        """
        return super(Cameras, self).show_image(camera_id, times)



"""Configure the logger based on a file or standard parameters."""
import json
import logging.config
import os


def configure():
    """
    Configures the logger.

    If a configuration file does not exist in the root user directory, then
    default configuration options will be written to file.  This file can be
    customized and will be read on the next configuration.

    """
    # Attempt to read SDK logging config file
    config_file = os.path.join(os.path.expanduser('~'),
                               'helios_logger_config.json')
    config = None
    if os.path.exists(config_file):
        try:
            with open(config_file, 'r') as f:
                config = json.load(f)
        except ValueError:
            config = None

    if config is None:
        # Default logging configuration.
        config = {'version': 1, 'disable_existing_loggers': 1}

        config['formatters'] = {}
        config['formatters']['simple'] = {'format': '%(asctime)s-%(levelname)s'
                                                    '-%(module)s-%(name)s'
                                                    '-%(funcName)s: %(message)s',
                                          'datefmt': '%H:%M:%S'}

        config['handlers'] = {}
        config['handlers']['console'] = {'class': 'logging.StreamHandler',
                                         'level': 'WARNING',
                                         'formatter': 'simple',
                                         'stream': 'ext://sys.stdout'}

        config['root'] = {'level': 'INFO',
                          'handlers': ['console']}

        # Write default configuration to file
        try:
            with open(config_file, 'w') as f:
                json.dump(config, f)
        except IOError:
            pass

    logging.config.dictConfig(config)



from setuptools import find_packages
from setuptools import setup

setup(name='helios-sdk',
      version='2.0.0',
      description='Use the Helios APIs in Python',
      author='Michael Bayer',
      author_email='mbayer@harris.com',
      url='https://github.com/harris-helios/helios-sdk-python',
      download_url='https://github.com/harris-helios/helios-sdk-python/archive/2.0.0.tar.gz',
      license='MIT',
      install_requires=['requests',
                        'numpy',
                        'Pillow',
                        'python_dateutil'],
      extras_require={
          'tests': ['pytest'],
      },
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      packages=find_packages())



"""
SDK for the Helios Observations API.

Methods are meant to represent the core functionality in the developer
documentation.  Some may have additional functionality for convenience.

"""
import logging
from contextlib import closing
from itertools import repeat
from multiprocessing.pool import ThreadPool

from helios.core import SDKCore, IndexMixin, ShowMixin, \
    DownloadImagesMixin, RequestManager
from helios.utilities import logging_utils


class Observations(DownloadImagesMixin, ShowMixin, IndexMixin, SDKCore):
    """
    The Observations API provides ground-truth data generated by the Helios
    analytics.

    """

    core_api = 'observations'
    max_threads = 32

    def __init__(self):
        self.request_manager = RequestManager(pool_maxsize=self.max_threads)
        self.logger = logging.getLogger(__name__)

    def index(self, **kwargs):
        """
        Return a list of observations matching the provided spatial, text, or
        metadata filters.

        The maximum skip value is 4000. If this is reached, truncated results
        will be returned. You will need to refine your query to avoid this.

        Args:
            **kwargs: Any keyword arguments found in the documentation.

        Returns:
             list: GeoJSON feature collections.

        """
        return super(Observations, self).index(**kwargs)

    def show(self, observation_id):
        """
        Return the attributes for a single observation.

        Args:
            observation_id (str): Observation ID.

        Returns:
            dict: GeoJSON feature.

        """
        return super(Observations, self).show(observation_id)

    @logging_utils.log_entrance_exit
    def preview(self, observation_ids, check_for_duds=True):
        """
        Return a preview image for an observation.

        This API call will attempt to filter out unusable images
        (e.g. full image text/logos, etc.) and will return the most recent
        image for the observation time period.

        Args:
            observation_ids (str or sequence of strs): list of observation IDs.
            check_for_duds (bool, optional): Flag to remove dud images from
                results. Defaults to True.

        Returns:
            sequence of strs: Image URLs.

        """
        # Force iterable
        if not isinstance(observation_ids, (list, tuple)):
            observation_ids = [observation_ids]
        n_obs = len(observation_ids)

        # Log entrance
        self.logger.info('Entering preview(%s observation_ids)', n_obs)

        # Get number of threads
        num_threads = min(self.max_threads, n_obs)

        # Process ids.
        if num_threads > 1:
            with closing(ThreadPool(num_threads)) as thread_pool:
                results = thread_pool.map(self.__preview_worker,
                                          zip(observation_ids, repeat(check_for_duds)))
        else:
            results = [self.__preview_worker((observation_ids[0], check_for_duds))]

        # Remove errors, if they exist
        results = [x for x in results if x != -1]

        # Check results
        n_successful = len(results)
        message = 'Leaving preview({} out of {} successful)'.format(n_successful, n_obs)

        if n_successful == 0:
            self.logger.error(message)
            return -1
        elif n_successful < n_obs:
            self.logger.warning(message)
        else:
            self.logger.info(message)

        return results

    def __preview_worker(self, args):
        observation_id, check_for_duds = args

        query_str = '{}/{}/{}/preview'.format(self.base_api_url,
                                              self.core_api,
                                              observation_id)

        try:
            resp = self.request_manager.get(query_str)
            redirect_url = resp.url[0:resp.url.index('?')]
        except Exception:
            return -1

        # Check header for dud statuses.
        if check_for_duds:
            try:
                # Redirect URLs do not use api credentials
                resp2 = self.request_manager.head(redirect_url,
                                                  use_api_cred=False)
            except Exception:
                return -1

            if self.check_headers_for_dud(resp2.headers):
                self.logger.info('preview query returned dud image: %s',
                                 query_str)
                return None

        return redirect_url



"""Manager for the authorization token required to access the Helios API."""
import json
import os

import requests


class TokenManager(object):
    """Manages API tokens for authentication."""

    token_expiration_threshold = 60  # minutes

    _token_file = os.path.join(os.path.expanduser('~'), '.helios_token')
    _auth_file = os.path.join(os.path.expanduser('~'), '.helios_auth')

    __default_api_url = r'https://api.helios.earth/v1/'

    def __init__(self):
        self.api_url = None
        self.token_url = None
        self.token = None

        self._key_id = None
        self._key_secret = None

        self.get_auth_credentials()

    def start_session(self):
        """
        Begin Helios session.

        Returns:
            tuple: The authentication token and API URL.

        """
        # Check for saved token first. If it doesn't exist then get a token.
        if os.path.exists(self._token_file):
            self.read_token()
            try:
                valid_token = self.verify_token()
                if not valid_token:
                    self.get_token()
            except Exception:
                raise
        else:
            self.get_token()

        return self.token, self.api_url

    def get_token(self):
        """Gets a fresh token and writes the token to file for reuse."""
        try:
            data = {'grant_type': 'client_credentials'}
            auth = (self._key_id, self._key_secret)
            resp = requests.post(self.token_url, data=data, auth=auth,
                                 verify=True)
            resp.raise_for_status()
        except requests.exceptions.HTTPError:
            token_url_http = 'http' + self.token_url.split('https')[1]
            data = {'grant_type': 'client_credentials'}
            auth = (self._key_id, self._key_secret)
            resp = requests.post(token_url_http, data=data, auth=auth,
                                 verify=True)

        # If the token cannot be acquired, raise exception.
        resp.raise_for_status()

        token_request = resp.json()
        self.token = {'name': 'Authorization',
                      'value': 'Bearer ' + token_request['access_token']}

        self.write_token()

    def read_token(self):
        """Read token from file."""
        with open(self._token_file, 'r') as token_file:
            self.token = json.load(token_file)

    def write_token(self):
        """Write token to file."""
        with open(self._token_file, 'w+') as token_file:
            json.dump(self.token, token_file)

    def delete_token(self):
        """Delete token file."""
        os.remove(self._token_file)

    def verify_token(self):
        """
        Verifies if the current token is still valid.

        If the token will expire in 60 minutes, then a new token will be
        acquired.

        """
        resp = requests.get(self.api_url + '/session',
                            headers={self.token['name']: self.token['value']},
                            verify=True)
        resp.raise_for_status()

        json_resp = resp.json()

        if json_resp['name'] is None:
            return False
        else:
            expiration_time = json_resp['expires_in'] / 60.0
            if expiration_time < self.token_expiration_threshold:
                self.get_token()
            return True

    def get_auth_credentials(self):
        if 'HELIOS_KEY_ID' in os.environ and 'HELIOS_KEY_SECRET' in os.environ:
            self._key_id = os.environ['HELIOS_KEY_ID']
            self._key_secret = os.environ['HELIOS_KEY_SECRET']

            # Check for API URL override in environment variables
            if 'HELIOS_API_URL' in os.environ:
                self.api_url = os.environ['HELIOS_API_URL']
                self.token_url = os.environ['HELIOS_API_URL'] + '/oauth/token'
            else:
                self.api_url = self.__default_api_url
                self.token_url = self.__default_api_url + '/oauth/token'

        elif os.path.exists(self._auth_file):
            with open(self._auth_file, 'r') as auth_file:
                data = json.load(auth_file)

            self._key_id = data['HELIOS_KEY_ID']
            self._key_secret = data['HELIOS_KEY_SECRET']

            # Check for API URL override in .helios_auth file
            if 'HELIOS_API_URL' in data:
                self.api_url = data['HELIOS_API_URL']
                self.token_url = data['HELIOS_API_URL'] + '/oauth/token'
            else:
                self.api_url = self.__default_api_url
                self.token_url = self.__default_api_url + '/oauth/token'
        else:
            raise Exception('No credentials could be found. Be sure to set '
                            'environment variables or .helios_auth file '
                            'correctly.')



"""Helper functions for JSON objects."""
import json


def read_json_file(json_file, **kwargs):
    """
    Read a json file.

    Args:
        json_file (str): Full path to JSON file.
        **kwargs: Any keyword argument from the json.load method.
    Returns:
        dict: JSON formatted dictionary.

    """
    with open(json_file, 'r') as f:
        return json.load(f, **kwargs)


def read_json_string(json_string, **kwargs):
    """
    Convert JSON formatted string to JSON.

    Args:
        json_string (str): JSON formatted string.
        **kwargs: Any keyword argument from the json.loads method.
    Returns:
        dict: JSON formatted dictionary.

    """
    return json.loads(json_string, **kwargs)


def write_json(json_dict, file_name, **kwargs):
    """
    Write JSON dictionary to file.

    Args:
        json_dict (dict): JSON formatted dictionary.
        file_name (str): Output file name.
        **kwargs: Any keyword argument from the json.dump method.
    Returns:
        None

    """
    with open(file_name, 'w+') as output_file:
        json.dump(json_dict, output_file, **kwargs)


def merge_json(data, keys):
    """
    Merge JSON fields into a single list.

    Keys can either be a single string or a list of strings signifying a chain
    of "keys" into the dictionary.

    Args:
        data (list): Dictionary to merge data from.
        keys (str or sequence of strs): A chain of keys into the dictionary
            to get to the field that will be merged.
    Returns:
        list: Merged values.

    """
    if not isinstance(keys, list):
        keys = [keys]
    for k in keys:
        data = _merge_digger(data, k)
    return data


def _merge_digger(data, key):
    merged_list = []
    if not isinstance(data, list):
        data = [data]
    for json_slice in data:
        temp = json_slice[key]
        if not isinstance(temp, list):
            temp = [temp]
        merged_list.extend(temp)
    return merged_list
